#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## VII Coupler
# Paul Dechamps

import fwk
from fwk.coloring import ccolors

from math import pi as M_PI
import numpy as np

class Coupler:
  def __init__(self, _iSolverAPI, vSolver) -> None:
    self.iSolverAPI = _iSolverAPI
    self.vSolver = vSolver
    
    self.vSolver.verbose = self.iSolverAPI.iSolver.verbose

    self.maxCouplIter = 150
    self.couplTol = 1e-4

    self.tms=fwk.Timers()
    self.resetInv = False
    pass

  def Run(self):
    self.tms['tot'].start()

    # Initilize meshes in c++ objects

    dragIter = np.zeros((0,2))

    self.nElmUpperPrev = 0
    couplIter = 0
    cdPrev = 0
    print('- AoA: {:<2.2f} Mach: {:<1.1f} Re: {:<2.1f}e6'
    .format(self.iSolverAPI.GetAlpha()*180/M_PI, self.iSolverAPI.GetMinf(), self.vSolver.GetRe()/1e6))
    print('{:>5s}| {:>10s} {:>10s} | {:>8s} {:>6s} {:>8s}'.format('It', 'Cl', 'Cd', 'xtrT', 'xtrB', 'Error'))
    print(' --------------------------------------------------------')
    while couplIter < self.maxCouplIter:

      # Run inviscid solver
      self.tms['inviscid'].start()
      if self.resetInv:
        self.iSolverAPI.ResetSolver()
      self.iSolverAPI.RunSolver()
      self.tms['inviscid'].stop()

      # Extract inviscid boundary
      self.tms['processing'].start()
      self.iSolverAPI.GetInvBC(self.vSolver, couplIter)
      self.tms['processing'].stop()

      # Run viscous solver
      self.tms['viscous'].start()
      exitCode = self.vSolver.Run(couplIter)
      self.tms['viscous'].stop()
      # print('Viscous ops terminated with exit code ', exitCode)

      self.tms['processing'].start()
      self.iSolverAPI.SetBlowingVel(self.vSolver)
      self.tms['processing'].stop()

      dragIter = np.vstack((dragIter, [self.iSolverAPI.GetCl(), self.vSolver.Cdt]))
      error = abs((self.vSolver.Cdt - cdPrev)/self.vSolver.Cdt) if self.vSolver.Cdt != 0 else 1
      
      if error  <= self.couplTol:

        print(ccolors.ANSI_GREEN,'{:>4.0f}| {:>10.5f} {:>10.5f} | {:>8.3f} {:>6.3f} {:>8.2f}\n'.format(couplIter,
        self.iSolverAPI.GetCl(), self.vSolver.Cdt, self.vSolver.Getxtr(0,0), self.vSolver.Getxtr(0,1), np.log10(error)),ccolors.ANSI_RESET)
        self.tms['tot'].stop()
        
        return dragIter
      
      cdPrev = self.vSolver.Cdt


      print('- {:>4.0f}| {:>10.5f} {:>10.5f} | {:>8.3f} {:>6.3f} {:>8.2f}'.format(couplIter,
      self.iSolverAPI.GetCl(), self.vSolver.Cdt, self.vSolver.Getxtr(0,0), self.vSolver.Getxtr(0,1), np.log10(error)))

      couplIter += 1
      """if couplIter == 5:
        quit()"""
      self.tms['processing'].stop()
    else:
      print(ccolors.ANSI_RED, 'Maximum number of {:<.0f} coupling iterations reached'.format(self.maxCouplIter), ccolors.ANSI_RESET)
      self.tms['tot'].stop()

      return dragIter


  def RunPolar(self, alphaSeq):
    
    import math

    self.iSolver.printOn = False
    self.vSolver.printOn = False

    coeffTable = np.zeros((3,len(alphaSeq)))
    coeffTable[0,:] = alphaSeq

    for ind, alpha in enumerate(alphaSeq):
      self.iSolver.pbl.update(alpha*M_PI/180)

      self.Run()

      # Collect Coeff

      coeffTable[1,ind] = self.iSolver.Cl
      coeffTable[2,ind] = self.vSolver.Cdt

    return coeffTable
