import numpy as np
from matplotlib import pyplot as plt
from fwk.coloring import ccolors

import dart.pyVII.vUtils as blxU

# 3D spline with Bézier curves
from scipy.interpolate import bisplrep
from scipy.interpolate import bisplev

from scipy.interpolate import interp1d

from scipy.interpolate import RBFInterpolator

import fwk



class Interpolator:
    def __init__(self, _dartSolver, bnd_wing, bnd_wake, _cfg):
        self.iSolver = _dartSolver

        self.bndWing = bnd_wing
        self.bndWake = bnd_wake
        self.nodesList, self.transformedCoord, self.wakeCoord, self.viscStruct, self.viscStruct_tr = self.__GetWing(bnd_wing, bnd_wake, _cfg)
        print('Sections created')
        self.config = _cfg

        self.xxPrev = [[np.zeros(0) for iReg in range(3)] for iSec in range(1)]
        #self.uePrev = [[np.zeros(len(self.viscStruct[iSec][iReg][:,0])) for iReg in range(3)] for iSec in range(len(self.viscStruct))]
        self.DeltaStarPrev = [[np.zeros(0
        ) for iReg in range(3)] for iSec in range(1)] 
        self.tms = fwk.Timers()

    def GetInvBC(self, vSolver, couplIter):
        self.tms['InvToVisc'].start()

        # Extract parameters from the inviscid solver
        # Particular care must be taken with space directions; In the inviscid, the airfoil is in the x,z plane.
        # While setting up the viscous solver, y and z must be inverted. 

        # Wing

        nElm = len(self.nodesList)
        U_wg, M_wg, Rho_wg = np.zeros((nElm, 3)), np.zeros(nElm), np.zeros(nElm)
        for iNode, n in enumerate(self.nodesList):
            U_wg[iNode,0] = self.iSolver.U[n.row][0]
            U_wg[iNode,1] = self.iSolver.U[n.row][1]
            U_wg[iNode,2] = self.iSolver.U[n.row][2]
            M_wg[iNode] = self.iSolver.M[n.row]
            Rho_wg[iNode] = self.iSolver.rho[n.row]
        
        # Wake

        nElm = len(self.bndWake.nodes)
        U_wk =np.zeros((nElm, 3))
        M_wk = np.zeros(nElm)
        Rho_wk = np.zeros(nElm)
        
        for iNode, n in enumerate(self.bndWake.nodes):
            U_wk[iNode,0] = self.iSolver.U[n.row][0]
            U_wk[iNode,1] = self.iSolver.U[n.row][1]
            U_wk[iNode,2] = self.iSolver.U[n.row][2]
            M_wk[iNode] = self.iSolver.M[n.row]
            Rho_wk[iNode] = self.iSolver.rho[n.row]

        if self.config['nDim'] == 3:  # x and y inverted in 3d
            U_wg[:, [1,2]] = U_wg[:,[2,1]]
            U_wk[:, [1,2]] = U_wk[:,[2,1]] 

        Ux = [self.__RbfToSections(self.transformedCoord[:,:2], U_wg[:,0], self.viscStruct_tr[0][0][:,:2]), self.__RbfToSections(self.wakeCoord[:, :1], U_wk[:,0], self.viscStruct[0][1][:,:1], 1)]
        Uy = [self.__RbfToSections(self.transformedCoord[:,:2], U_wg[:,1], self.viscStruct_tr[0][0][:,:2]), self.__RbfToSections(self.wakeCoord[:, :1], U_wk[:,1], self.viscStruct[0][1][:,:1], 1)]
        Uz = [self.__RbfToSections(self.transformedCoord[:,:2], U_wg[:,2], self.viscStruct_tr[0][0][:,:2]), self.__RbfToSections(self.wakeCoord[:, :1], U_wk[:,2], self.viscStruct[0][1][:,:1], 1)]

        Me = [self.__RbfToSections(self.transformedCoord[:,:2], M_wg, self.viscStruct_tr[0][0][:,:2]), self.__RbfToSections(self.wakeCoord[:, :1], M_wk, self.viscStruct[0][1][:,:1], 1)]
        Rho = [self.__RbfToSections(self.transformedCoord[:,:2], Rho_wg, self.viscStruct_tr[0][0][:,:2]), self.__RbfToSections(self.wakeCoord[:, :1], Rho_wk, self.viscStruct[0][1][:,:1], 1)]

        # Find stagnation point
        for iSec in range(len(self.viscStruct)):

            if couplIter == 0:
                self.stgPt = np.argmin(Ux[0]**2+Uy[0]**2)
            
                xUp, xLw = self.__Remesh(self.viscStruct[iSec][0][:,0], self.stgPt)
                yUp, yLw = self.__Remesh(self.viscStruct[iSec][0][:,1], self.stgPt)
                zUp, zLw = self.__Remesh(self.viscStruct[iSec][0][:,2], self.stgPt)

                vSolver.SetGlobMesh(iSec, 0, xUp, yUp, zUp)
                vSolver.SetGlobMesh(iSec, 1, xLw, yLw, zLw)
                vSolver.SetGlobMesh(iSec, 2, self.viscStruct[iSec][1][:,0],
                                            self.viscStruct[iSec][1][:,1],
                                            self.viscStruct[iSec][1][:,2])

                self.DeltaStarPrev[0][0] = np.zeros(len(xUp))
                self.DeltaStarPrev[0][1] = np.zeros(len(xLw))
                self.DeltaStarPrev[0][2] = np.zeros(len(self.viscStruct[iSec][1][:,0]))
                
                self.xxPrev[0][0] = np.zeros(len(xUp))
                self.xxPrev[0][1] = np.zeros(len(xLw))
                self.xxPrev[0][2] = np.zeros(len(self.viscStruct[iSec][1][:,0]))
            
            UxUp, UxLw = self.__Remesh(Ux[0], self.stgPt)
            UyUp, UyLw = self.__Remesh(Uy[0], self.stgPt)
            UzUp, UzLw = self.__Remesh(Uz[0], self.stgPt)

            vSolver.SetVelocities(iSec, 0, UxUp, UyUp, UzUp)
            vSolver.SetVelocities(iSec, 1, UxLw, UyLw, UzLw)
            vSolver.SetVelocities(iSec, 2, Ux[1], Uy[1], Uz[1])

            MeUp, MeLw = self.__Remesh(Me[0], self.stgPt)

            vSolver.SetMe(iSec, 0, MeUp)
            vSolver.SetMe(iSec, 1, MeLw)
            vSolver.SetMe(iSec, 2, Me[1])

            RhoUp, RhoLw = self.__Remesh(Rho[0], self.stgPt)

            vSolver.SetRhoe(iSec, 0, RhoUp)
            vSolver.SetRhoe(iSec, 1, RhoLw)
            vSolver.SetRhoe(iSec, 2, Rho[1])

            #DeltaStarPrevUp, DeltaStarPrevLw = self.__Remesh(self.DeltaStarPrev[iSec][0], self.stgPt)
            vSolver.SetDeltaStarExt(iSec, 0, self.DeltaStarPrev[0][0])
            vSolver.SetDeltaStarExt(iSec, 1, self.DeltaStarPrev[0][1])
            vSolver.SetDeltaStarExt(iSec, 2, self.DeltaStarPrev[0][2])

            #xxPrevUp, xxPrevLw = self.__Remesh(self.xxPrev[iSec][0], self.stgPt)
            vSolver.SetxxExt(iSec, 0, self.xxPrev[0][0])
            vSolver.SetxxExt(iSec, 1, self.xxPrev[0][1])
            vSolver.SetxxExt(iSec, 2, self.xxPrev[0][2])

            """uePrevUp, uePrevLw = self.__Remesh(self.uePrev[iSec][0], self.stgPt)
            vSolver.SetUeOld(iSec, 0, uePrevUp)
            vSolver.SetUeOld(iSec, 1, uePrevLw)
            vSolver.SetUeOld(iSec, 2, self.uePrev[iSec][1])"""
            
            

    def SetBlowingVel(self, vSolver):
        BlowingVel = [[np.zeros(0) for iReg in range(3)] for iSec in range(len(self.viscStruct))]
        for iSec in range(len(self.viscStruct)):
            for iRegion in range(3):
                BlowingVel[iSec][iRegion] = np.asarray(vSolver.ExtractBlowingVel(iSec, iRegion))
                self.DeltaStarPrev[iSec][iRegion] = np.asarray(vSolver.ExtractDeltaStar(iSec, iRegion))
                self.xxPrev[iSec][iRegion] = np.asarray(vSolver.Extractxx(iSec, iRegion))

        if self.config['nDim'] == 2:
            BlwWing = np.concatenate((BlowingVel[0][0][::-1], BlowingVel[0][1]))
            BlwWake = BlowingVel[0][2]
            """self.DeltaStarPrev[iSec][0] = np.concatenate((np.asarray(vSolver.ExtractDeltaStar(iSec, 0))[::-1], np.asarray(vSolver.ExtractDeltaStar(iSec, 1))[1:]), axis=None)
            self.DeltaStarPrev[iSec][1] = np.asarray(vSolver.ExtractDeltaStar(iSec, 2))
            self.xxPrev[iSec][0] = np.concatenate((np.asarray(vSolver.Extractxx(iSec, 0))[::-1], np.asarray(vSolver.Extractxx(iSec, 1))[1:]), axis=None)
            self.xxPrev[iSec][1] = np.asarray(vSolver.Extractxx(iSec, 2))
            self.uePrev[iSec][0] = np.concatenate((np.asarray(vSolver.ExtractUe(iSec, 0))[::-1], np.asarray(vSolver.ExtractUe(iSec, 1))[1:]), axis=None)
            self.uePrev[iSec][1] = np.asarray(vSolver.ExtractUe(iSec, 2))"""
            
            meany = (self.viscStruct[0][0][np.argmin(self.viscStruct[0][0][:,0]), 1] - self.viscStruct[0][0][np.argmax(self.viscStruct[0][0][:,0]), 1]) / 2

            # Compute cell centers
            viscCgWing = np.zeros((len(self.viscStruct[0][0][:,0]) - 1, 3))
            for iElm in range(len(self.viscStruct[0][0][:,0]) - 1):
                viscCgWing[iElm, 0] = self.viscStruct[0][0][iElm, 0] + self.viscStruct[0][0][iElm + 1, 0]
                viscCgWing[iElm, 1] = self.viscStruct[0][0][iElm, 1] + self.viscStruct[0][0][iElm + 1, 1]
                viscCgWing[iElm, 2] = self.viscStruct[0][0][iElm, 2] + self.viscStruct[0][0][iElm + 1, 2]
            viscCgWing/=2
            viscCgWing[viscCgWing[:,1]<meany, 0] *= -1

            viscCgWake = np.zeros((len(self.viscStruct[0][1][:,0]) - 1, 3))
            for iElm in range(len(self.viscStruct[0][1][:,0]) - 1):
                viscCgWake[iElm, 0] = self.viscStruct[0][1][iElm, 0] + self.viscStruct[0][1][iElm + 1, 0]
                viscCgWake[iElm, 1] = self.viscStruct[0][1][iElm, 1] + self.viscStruct[0][1][iElm + 1, 1]
                viscCgWake[iElm, 2] = self.viscStruct[0][1][iElm, 2] + self.viscStruct[0][1][iElm + 1, 2]
            viscCgWake/=2

            invCgWing = np.zeros((len(self.bndWing.tag.elems), 3))
            for iElm, e in enumerate(self.bndWing.tag.elems):
                for n in e.nodes:
                    invCgWing[iElm, 0] +=n.pos[0]
                    invCgWing[iElm, 1] +=n.pos[1]
                    invCgWing[iElm, 2] +=n.pos[2]
                invCgWing[iElm, :] /= e.nodes.size()
            
            invCgWing[invCgWing[:,1]<meany, 0] *= -1

            # Wake 

            invCgWake = np.zeros((len(self.bndWake.tag.elems), 3))
            for iElm, e in enumerate(self.bndWake.tag.elems):
                for n in e.nodes:
                    invCgWake[iElm, 0] +=n.pos[0]
                    invCgWake[iElm, 1] +=n.pos[1]
                    invCgWake[iElm, 2] +=n.pos[2]
                invCgWake[iElm, :] /= e.nodes.size()
            
            if self.config['nDim'] == 3:
                invCgWing[:,[1,2]] = invCgWing[:,[2,1]]
                invCgWake[:,[1,2]] = invCgWake[:,[2,1]]

        self.tms['Interpolation'].start()

        mappedBlw = [self.__RbfToSections(viscCgWing[:, :2], BlwWing, invCgWing[:,:2]), self.__RbfToSections(viscCgWake[:, :1], BlwWake, invCgWake[:, :1])]

        #mappedBlwWing = RBFInterpolator(viscCgWing[:,:2], BlwWing, neighbors=self.config['neighbors'], kernel=self.config['rbftype'], smoothing=self.config['smoothing'], degree=self.config['degree'])(invCgWing[:,:2])
        #mappedBlwWake = RBFInterpolator(viscCgWake[:, :2], BlwWake, neighbors=self.config['neighbors'], kernel=self.config['rbftype'], smoothing=self.config['smoothing'], degree=self.config['degree'])(invCgWake[:, :2])

        sln = blxU.GetSolution(0, vSolver)

        self.tms['Interpolation'].stop()
        for iElm in range(len(self.bndWing.tag.elems)):
            self.bndWing.setU(iElm, mappedBlw[0][iElm])
        for iElm in range(len(self.bndWake.tag.elems)):
            self.bndWake.setU(iElm, mappedBlw[1][iElm])

        


    def RunSolver(self):
        self.iSolver.run()

    def ResetSolver(self):
        self.iSolver.reset()

    def GetCp(self,bnd):
        import dart.utils as floU

        Cp = floU.extract(bnd, self.iSolver.Cp)
        return Cp

    def GetAlpha(self):
        return self.iSolver.pbl.alpha

    def GetMinf(self):
        return self.iSolver.pbl.M_inf
    
    def GetCl(self):
        return self.iSolver.Cl
    
    def GetCm(self):
        return self.iSolver.Cm

    def __Remesh(self, vec, idx_stag):
        xUp = vec[:idx_stag+1]
        xUp = xUp[::-1]

        xLw = vec[idx_stag:]
        return xUp, xLw

    def __RbfToSections(self, x, var, xs, wake=0):

        if wake==0:
            return RBFInterpolator(x, var, neighbors=self.config['neighbors'], kernel=self.config['rbftype'], smoothing=self.config['smoothing'], degree=self.config['degree'])(xs)
        else:
            return RBFInterpolator(x, var, neighbors=self.config['neighbors'], kernel=self.config['rbftype'], smoothing=self.config['smoothing'], degree=self.config['degree'])(xs)

    def __GetWing(self, bnd_wing, bnd_wake, config):

        if config['nDim'] == 2 and config['nSections'] != 1:
            raise RuntimeError('Cannot create more than 1 section on 2D geometry. Specified:', config['nSections'])

        viscStruct = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]
        viscStruct_tr = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]

        # Create node list
        nodesCoord = np.zeros((0, 3))
        for n in bnd_wing.nodes:
            nodesCoord = np.vstack((nodesCoord, [n.pos[0], n.pos[1], n.pos[2]]))

        if config['nDim'] == 2:
            spanDistri = np.zeros(1)
            spanDistri[0] = nodesCoord[0,2] # Span direction is the z direction
        
        elif config['nDim'] == 3:
            nodesCoord[:, [1,2]] = nodesCoord[:, [2, 1]]
            spanDistri = np.linspace(min(nodesCoord[:,2]), max(nodesCoord[:,2]), config['nSections'])
            spanDistri[0] +=0.1
            spanDistri[-1] -=0.1

        else:
            raise RuntimeError("Number of dimensions not recognized", config['nDim'])

        wingSpan = max(nodesCoord[:,2]) - min(nodesCoord[:,2])

        for iSec, z in enumerate(spanDistri):
            
            portion = nodesCoord[abs(nodesCoord[:,2] - z) <= 0.01 * wingSpan,:]

            le_coord = portion[portion[:,0] == np.min((portion[:,0])), :][0]
            te_coord = portion[portion[:,0] == np.max((portion[:,0])), :][0]

            # Find te elements
            te_elems = []
            for e in bnd_wing.tag.elems:
                for n in e.nodes:
                    if n.pos[0] == te_coord[0]:
                        te_elems.append(e)
                        break
            
            # Cg of te elements
            cg_teElems = np.zeros((len(te_elems), 3))
            for iElm, eTe in enumerate(te_elems):
                for n in eTe.nodes:
                    cg_teElems[iElm, 0] += n.pos[0]
                    cg_teElems[iElm, 1] += n.pos[1]
                    cg_teElems[iElm, 2] += n.pos[2]
            
            # Sort upper side in 0 and lower side in 1
            if cg_teElems[0, 1] < cg_teElems[1, 1]:
                save = te_elems[0].copy()
                te_elems[0] = te_elems[1]
                te_elems[1] = save
            
            # Find te_nodes (sorted [upper, lower])
            te_nodes = [None, None]
            for i in range(len(te_elems)):
                for n in te_elems[i].nodes:
                    if n.pos[0] == te_coord[0]:
                        te_nodes[i] = n
                        break
            
            # Find le_node
            for n in bnd_wing.nodes:
                if n.pos[0] == le_coord[0]:
                    le_node = n
                    break
                

            meany = (le_coord[1] + te_coord[1]) / 2

            transformedCoord = np.zeros((0, 3))
            nodesList = []

            for iNode, n in enumerate(bnd_wing.nodes):
                if n.pos[0] != le_coord[0] and n.pos[0] != te_coord[0]:
                    if n.pos[1] > meany:
                        transformedCoord = np.vstack((transformedCoord, [n.pos[0], n.pos[1], n.pos[2]]))
                        nodesList.append(n)

                    elif n.pos[1] < meany:
                        transformedCoord = np.vstack((transformedCoord, [-n.pos[0], n.pos[1], n.pos[2]]))
                        nodesList.append(n)

            transformedCoord = np.vstack((transformedCoord, le_coord))
            nodesList.append(le_node)
            transformedCoord = np.vstack((transformedCoord, [te_nodes[0].pos[0], te_nodes[0].pos[1], te_nodes[0].pos[2]]))
            nodesList.append(te_nodes[0])
            transformedCoord = np.vstack((transformedCoord, [-te_nodes[1].pos[0], te_nodes[1].pos[1], te_nodes[1].pos[2]]))
            nodesList.append(te_nodes[1])

            wakeCoord = np.zeros((len(bnd_wake.nodes), 3))
            for iNode, n in enumerate(bnd_wake.nodes):
                wakeCoord[iNode, 0] = n.pos[0]
                wakeCoord[iNode, 1] = n.pos[1]
                wakeCoord[iNode, 2] = n.pos[2]
            
            chordLength = max(portion[:, 0]) - min(portion[:, 0])
            xDistri = le_coord[0] + 1/2*(1 + np.cos(np.linspace(0, np.pi, config['nPoints'][0])))*chordLength

            viscStruct[iSec][0][:,0] = np.concatenate((xDistri[:-1], xDistri[::-1]), axis=None)
            viscStruct[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            viscStruct_tr[iSec][0][:,0] = np.concatenate((xDistri[:-1], -xDistri[::-1]), axis=None).copy()
            viscStruct_tr[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            # Interpolation in the transformed space.
            viscStruct_tr[iSec][0][:,1] = interp1d(transformedCoord[:,0], transformedCoord[:,1])(viscStruct_tr[iSec][0][:,0])
            viscStruct[iSec][0][:,1] = viscStruct_tr[iSec][0][:,1].copy()

            # Wake
            wakeLength = np.max(wakeCoord[:,0]) - np.max(viscStruct[iSec][0][:,0])
            viscStruct[iSec][1][:,0] = np.max(viscStruct[iSec][0][:,0]) + (np.max(viscStruct[iSec][0][:,0]) + np.cos(np.linspace(np.pi, np.pi/2, config['nPoints'][1]))) * wakeLength
            viscStruct[iSec][1][:,2] = z*np.ones(len(viscStruct[iSec][1][:,0]))
            viscStruct[iSec][1][:,1] = interp1d(wakeCoord[:,0], wakeCoord[:,1])(viscStruct[iSec][1][:,0])

            

            return nodesList, transformedCoord, wakeCoord, viscStruct, viscStruct_tr

            """for iSec, z in enumerate(spanDistri):

            # Create x distribution at the corresponding span position
            minX = min(upperCoord[abs(upperCoord[:,2]-z)<=config['eps'], 0])
            chordLength = max(upperWing[abs(upperWing[:,2]-z)<=config['eps'], 0]) - minX
            xDistri = minX + 1/2*(1 + np.cos(np.linspace(0, np.pi, config['nPoints'][0])))*chordLength
            nElmSide = len(xDistri)
            viscStruct[iSec][0][:,0] = np.concatenate((xDistri, xDistri[::-1]), axis=None)
            viscStruct[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            wakeLength = np.max(wakeCoord[:,0]) - np.max(viscStruct[iSec][0][:,0])
            viscStruct[iSec][1][:,0] = np.max(viscStruct[iSec][0][:,0]) + (np.max(viscStruct[iSec][0][:,0]) + np.cos(np.linspace(np.pi, np.pi/2, config['nPoints'][1]))) * wakeLength
            viscStruct[iSec][1][:,2] = z*np.ones(len(viscStruct[iSec][1][:,0]))

            if config['nDim'] == 2:
                transformedViscCoord = np.zeros((np.shape(viscStruct[iSec][0])))
                transformedCoord
                transformedViscCoord[:, 1] = interp1d(transformedCoord[:,0], transformedCoord[:,1])(transformedCoord[:,0])
                viscStruct[iSec][0][:nElmSide,1] = interp1d(upper[:,0], upper[:,1])(viscStruct[iSec][0][:nElmSide,0])
                viscStruct[iSec][0][nElmSide:,1] = interp1d(lower[:,0], lower[:,1])(viscStruct[iSec][0][nElmSide:,0])
                viscStruct[iSec][0] = np.delete(viscStruct[iSec][0], nElmSide, axis=0)

                viscStruct[iSec][1][:,1] = interp1d(invStruct[1][:,0], invStruct[1][:,1])(viscStruct[iSec][1][:,0])

                plt.plot(viscStruct[iSec][0][:,0], viscStruct[iSec][0][:,1], '-')
                plt.plot(viscStruct[iSec][0][:nElmSide,0], viscStruct[iSec][0][:nElmSide,1], 'x')
                plt.plot(viscStruct[iSec][0][nElmSide:,0], viscStruct[iSec][0][nElmSide:,1], 'o')
                plt.plot(viscStruct[iSec][1][:,0], np.zeros(len(viscStruct[iSec][1][:,0])), 'x-')
                plt.xlim([-0.5,2])
                plt.ylim([-0.5, 0.5])
                plt.show()

            elif config['nDim'] == 3:
                nPointsMin = (config['k'][0]+1) * (config['k'][1]+1) # See Scipy bisplrep doc

                wingSpan = max(invStruct[0][:,2]) - min(invStruct[0][:,2])
                portion = invStruct[0][abs(invStruct[0][:,2] - z) < 0.01 * wingSpan,:]

                portionUpper, portionLower = self.__SeparateArf(portion[:,0], portion[:,1], portion[:,2])
                if len(portionUpper)<=nPointsMin or len(portionLower)<=nPointsMin:
                    print(ccolors.ANSI_YELLOW,'dart::Interpolator Warning: Found {:.0f} points within 1% of spanwise position z={:.2f}. Trying 10%. '.format(
                        len(portion), z),ccolors.ANSI_RESET)
                    portion = invStruct[0][abs(invStruct[0][:,2] - z) < 0.1 * wingSpan,:]
                    if len(portion)<=nPointsMin:
                        print(ccolors.ANSI_RED, 'Failed. Continuing\n', ccolors.ANSI_RESET)
                    else:
                        print(ccolors.ANSI_GREEN, 'Sucess.\n', ccolors.ANSI_RESET)

                # Upper
                sWing = (len(portionUpper[:,0])-np.sqrt(2*len(portionUpper[:,0])) + len(portionUpper[:,0])+np.sqrt(2*len(portionUpper[:,0])))/2
                tckWingUp = bisplrep(portionUpper[:,0], portionUpper[:,2], portionUpper[:,1], kx=config['k'][0], ky=config['k'][1], s=sWing)
                for iPoint in range(nElmSide):
                    viscStruct[iSec][0][iPoint,1] = bisplev(viscStruct[iSec][0][iPoint,0], viscStruct[iSec][0][iPoint,2], tckWingUp)
                
                # Lower
                sWing = (len(portionLower[:,0])-np.sqrt(2*len(portionLower[:,0])) + len(portionLower[:,0])+np.sqrt(2*len(portionLower[:,0])))/2
                tckWingUp = bisplrep(portionLower[:,0], portionLower[:,2], portionLower[:,1], kx=config['k'][0], ky=config['k'][1], s=sWing)
                for iPoint in range(nElmSide, len(viscStruct[iSec][0][:,0])):
                    viscStruct[iSec][0][iPoint,1] = bisplev(viscStruct[iSec][0][iPoint,0], viscStruct[iSec][0][iPoint,2], tckWingUp)
                
                invStruct[1] = invStruct[1][invStruct[1][:, 0].argsort()]
                fig = plt.figure()
                ax = fig.add_subplot(projection='3d')
                ax.scatter(invStruct[1][:,0], invStruct[1][:,1], invStruct[1][:,2], s=0.2)
                ax.set_ylim([-0.5, 0.5])
                ax.set_xlabel('x')
                ax.set_ylabel('y')
                ax.set_zlabel('z')
                plt.show()
                # Wake
                sWake = (len(invStruct[1][:,0])-np.sqrt(2*len(invStruct[1][:,0])) + len(invStruct[1][:,0])+np.sqrt(2*len(invStruct[1][:,0])))/2
                tckWake = bisplrep(invStruct[1][:,0], invStruct[1][:,2], invStruct[1][:,1], kx=config['k'][0], ky=config['k'][1], s=sWake)
                for iPoint in range(len(viscStruct[iSec][1][:,0])):
                    #viscStruct[iSec][1][iPoint,1] = bisplev(viscStruct[iSec][1][iPoint,0], viscStruct[iSec][1][iPoint,2], tckWake)
                    viscStruct[iSec][1][iPoint,1] = 0

        if config['nDim'] ==3:
            fig = plt.figure()
            ax = fig.add_subplot(projection='3d')
            ax.scatter(invStruct[0][:,0], invStruct[0][:,1], invStruct[0][:,2], s=0.2)
            for iSec in range(len(viscStruct)):
                ax.scatter(viscStruct[iSec][0][:,0], viscStruct[iSec][0][:,1], viscStruct[iSec][0][:,2], s=0.9, color='red')
            ax.set_ylim([-0.5, 0.5])
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            plt.show()
        return invStruct, viscStruct"""

    def __SeparateArf(self, xInp, yInp, zInp=None):
        if zInp is None:
            zInp = np.zeros(len(xInp))
        if len(xInp) != len(yInp) or len(xInp) != len(zInp):
            raise RuntimeError ('dart::Interpolator::__SeparateArf Missmatch between intput vector sizes.')

        Q=np.zeros((len(xInp), 3))
        Q[:,0] = xInp.copy()
        Q[:,1] = yInp.copy()
        Q[:,2] = zInp.copy()

        Q = Q[Q[:, 0].argsort()]

        le = Q[0,:]
        te = Q[-1,:]
        ymean = (le[1] + te[1]) / 2

        upper, lower, save = le, le, le

        flag1 = 0
        iPoint = 0
        for i in range(len(Q[:,0])):
            if Q[iPoint,1] == 0.0:
                if flag1 == 1:
                    Q=np.delete(Q, iPoint, axis = 0)
                    iPoint-=1
                else:
                    flag1 = 1
            iPoint +=1

        for iPoint in range(1, len(Q[:,0])):
            if Q[iPoint,1] > ymean:
                upper = np.vstack((upper, Q[iPoint,:]))
            elif Q[iPoint, 1] < ymean:
                lower = np.vstack((lower, Q[iPoint,:]))
            else:
                save = np.vstack((save, Q[iPoint, :]))
        
        _, idx = np.unique(save, return_index=True, axis=0)
        save=save[np.sort(idx)]

        #save = np.delete(save, 0, axis=0)

        
        try:
            if len(save[:,0])>0:
                upper = np.vstack((upper, save))
                lower = np.vstack((lower, save))
        except:
            pass

        upper = np.vstack((upper, te))
        lower = np.vstack((lower, te))

        return upper, lower
