#include "wTimeSolver.h"
#include "wBLRegion.h"
#include "wViscFluxes.h"
#include "wInvBnd.h"
#include <Eigen/Dense>
#include <iostream>
#include <vector>
#include <cmath>

using namespace Eigen;
using namespace tbox;
using namespace dart;

TimeSolver::TimeSolver(double _CFL0, double _Minf, double Re, unsigned long _maxIt, double _tol, unsigned int _itFrozenJac)
{

    CFL0 = _CFL0;
    maxIt = _maxIt;
    tol = _tol;
    itFrozenJac0 = _itFrozenJac;
    Minf = std::max(_Minf, 0.1);
    initSln = true;

    SysMatrix = new ViscFluxes(Re);

} // end TimeSolver

TimeSolver::~TimeSolver()
{
    delete SysMatrix;
    std::cout << "~TimeSolver()\n";
} // end ~TimeSolver

void TimeSolver::InitialCondition(size_t iPoint, BLRegion *bl)
{
    unsigned int nVar = bl->GetnVar();
    if (initSln)
    {

        bl->U[iPoint * nVar + 0] = bl->U[(iPoint - 1) * nVar + 0];
        bl->U[iPoint * nVar + 1] = bl->U[(iPoint - 1) * nVar + 1];
        bl->U[iPoint * nVar + 2] = bl->U[(iPoint - 1) * nVar + 2];
        bl->U[iPoint * nVar + 3] = bl->invBnd->GetVt(iPoint);
        bl->U[iPoint * nVar + 4] = bl->U[(iPoint - 1) * nVar + 4];

        bl->Ret[iPoint] = bl->Ret[iPoint - 1];
        bl->Cd[iPoint] = bl->Cd[iPoint - 1];
        bl->Cf[iPoint] = bl->Cf[iPoint - 1];
        bl->Hstar[iPoint] = bl->Hstar[iPoint - 1];
        bl->Hstar2[iPoint] = bl->Hstar2[iPoint - 1];
        bl->Hk[iPoint] = bl->Hk[iPoint - 1];
        bl->Cteq[iPoint] = bl->Cteq[iPoint - 1];
        bl->Us[iPoint] = bl->Us[iPoint - 1];
        bl->Delta[iPoint] = bl->Delta[iPoint - 1];
        bl->DeltaStar[iPoint] = bl->DeltaStar[iPoint - 1];
    }
    if (bl->Regime[iPoint] == 1 && bl->U[iPoint * nVar + 4] <= 0)
    {
        bl->U[iPoint * nVar + 4] = 0.03;
    } // End if

} // End InitialCondition

int TimeSolver::Integration(size_t iPoint, BLRegion *bl)
{

    unsigned int nVar = bl->GetnVar();

    /* Save initial condition */

    std::vector<double> Sln0(nVar, 0.);
    for (size_t i = 0; i < nVar; ++i)
    {
        Sln0[i] = bl->U[iPoint * nVar + i];
    } // End for

    /* Initialize time integration variables */

    double dx = bl->mesh->GetLoc(iPoint) - bl->mesh->GetLoc(iPoint - 1);

    /* Initial time step */

    double CFL = CFL0;
    unsigned int itFrozenJac = itFrozenJac0;
    double timeStep = SetTimeStep(CFL, dx, bl->invBnd->GetVt(iPoint));
    //MatrixXd IMatrix(5,5);
    MatrixXd IMatrix(5,5);
    IMatrix = MatrixXd::Identity(5,5) / timeStep;

    /* Initial system residuals */

    VectorXd SysRes = SysMatrix->ComputeFluxes(iPoint, bl);
    double normSysRes0 = SysRes.norm();
    double normSysRes = normSysRes0;

    MatrixXd JacMatrix = MatrixXd::Zero(5, 5);
    VectorXd slnIncr = VectorXd::Zero(5);

    unsigned int nErrors = 0;   // Number of errors encountered
    unsigned int innerIters = 0; // Inner (non-linear) iterations

    /* for (int i=0; i<1e10; ++i)
    {
        double a = 1+1;
    } */
    while (innerIters < maxIt)
    {
        /* if (bl->name=="upper" && iPoint == 1 && innerIters == 0)
        {
            std::cout << "it " << innerIters << " res " << std::log10(normSysRes/normSysRes0) << std::endl;
            bl->PrintSolution(0);
        } */

        if (innerIters % itFrozenJac == 0)
        {
            JacMatrix = SysMatrix->ComputeJacobian(iPoint, bl, SysRes, 1e-8);
        }

        slnIncr = (JacMatrix + IMatrix).partialPivLu().solve(-SysRes);

        for (size_t k = 0; k < nVar; ++k)
        {
            bl->U[iPoint * nVar + k] += slnIncr(k);
        }
        bl->U[iPoint * nVar + 0] = std::max(bl->U[iPoint * nVar + 0], 1e-6);
        bl->U[iPoint * nVar + 1] = std::max(bl->U[iPoint * nVar + 1], 1.0005);

        SysRes = SysMatrix->ComputeFluxes(iPoint, bl);
        normSysRes = (SysRes + slnIncr / timeStep).norm();

        if (std::isnan(normSysRes))
        {
            ResetSolution(iPoint, bl, Sln0, true);
            SysRes = SysMatrix->ComputeFluxes(iPoint, bl);
            CFL = 1.0;
            timeStep = SetTimeStep(CFL, dx, bl->invBnd->GetVt(iPoint));
            IMatrix = MatrixXd::Identity(5,5) / timeStep;
            normSysRes = (SysRes + slnIncr / timeStep).norm();
            itFrozenJac = 1;
        }

        if (normSysRes <= tol * normSysRes0)
        {
            return 0; /* Successfull exit */
        }

        /* CFL adaptation. */

        CFL = std::max(CFL0 * pow(normSysRes0 / normSysRes, 0.7), 0.1);
        timeStep = SetTimeStep(CFL, dx, bl->invBnd->GetVt(iPoint));
        IMatrix = MatrixXd::Identity(5,5) / timeStep;

        innerIters++;
    } // End while (innerIters < maxIt)
    if (std::isnan(normSysRes) || normSysRes/normSysRes0 > 1e-3)
    {
        ResetSolution(iPoint, bl, Sln0);
    }
    return innerIters;
} // End Integration

double TimeSolver::SetTimeStep(double CFL, double dx, double invVel) const
{
    /* Speed of sound. */
    double gradU2 = (invVel * invVel);
    double soundSpeed = ComputeSoundSpeed(gradU2);

    /* Velocity of the fastest travelling wave */
    double advVel = soundSpeed + invVel;

    /* Time step computation */
    return CFL * dx / advVel;
}

double TimeSolver::ComputeSoundSpeed(double gradU2) const
{
    return sqrt(1 / (Minf * Minf) + 0.5 * (gamma - 1) * (1 - gradU2));
}

void TimeSolver::ResetSolution(size_t iPoint, BLRegion *bl, std::vector<double> Sln0, bool usePrevPoint)
{

    unsigned int nVar = bl->GetnVar();

    /* Reset solution */

    if (usePrevPoint)
    {
        for (size_t k = 0; k < nVar; ++k)
        {
            bl->U[iPoint * nVar + k] = bl->U[(iPoint - 1) * nVar + k];
        } // End for
    }     // End if (usePrevPoint)
    else
    {
        for (size_t k = 0; k < nVar; ++k)
        {
            bl->U[iPoint * nVar + k] = Sln0[k];
        } // End for
    }     // End else (usePrevPoint)

    /* Reset closures */

    bl->Ret[iPoint] = std::max( bl->U[iPoint * nVar + 3] *  bl->U[iPoint * nVar + 0] * SysMatrix->Re, 1.0);  /* Reynolds number based on theta */
    bl->DeltaStar[iPoint] = bl->U[iPoint * nVar + 0] * bl->U[iPoint * nVar + 1]; /* Displacement thickness */

    if (bl->Regime[iPoint] == 0)
    {
        std::vector<double> lamParam = bl->closSolver->LaminarClosures(bl->U[iPoint * nVar + 0], bl->U[iPoint * nVar + 1], bl->Ret[iPoint], bl->invBnd->GetMe(iPoint), bl->name);
        bl->Hstar[iPoint] = lamParam[0];
        bl->Hstar2[iPoint] = lamParam[1];
        bl->Hk[iPoint] = lamParam[2];
        bl->Cf[iPoint] = lamParam[3];
        bl->Cd[iPoint] = lamParam[4];
        bl->Delta[iPoint] = lamParam[5];
        bl->Cteq[iPoint] = lamParam[6];
        bl->Us[iPoint] = lamParam[7];
    } // End if
    else if (bl->Regime[iPoint] == 1)
    {
        std::vector<double> turbParam = bl->closSolver->TurbulentClosures(bl->U[iPoint * nVar + 0], bl->U[iPoint * nVar + 1], bl->U[iPoint * nVar + 4], bl->Ret[iPoint], bl->invBnd->GetMe(iPoint), bl->name);
        bl->Hstar[iPoint] = turbParam[0];
        bl->Hstar2[iPoint] = turbParam[1];
        bl->Hk[iPoint] = turbParam[2];
        bl->Cf[iPoint] = turbParam[3];
        bl->Cd[iPoint] = turbParam[4];
        bl->Delta[iPoint] = turbParam[5];
        bl->Cteq[iPoint] = turbParam[6];
        bl->Us[iPoint] = turbParam[7];
    } // End else if

} // End ResetSolution